package type3;

import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import auxiliary.ListOfListContains;
import co_reachable.CoReachability;
import reachable.Reachability;

public class Type3GrammarGenerator {

	Random random = new Random();

	private Map<String, List<List<String>>> genType3GrammarRuleSet(String[] vars, int alphaSize, int maxConclusios,
			String s) {
		Map<String, List<List<String>>> rules = new LinkedHashMap<>();
		List<List<String>> conclusios0 = genConclusios(alphaSize, vars, maxConclusios, true);
		rules.put(s, conclusios0);
		for (int i = 1; i < vars.length; i++) {
			List<List<String>> conclusios1 = genConclusios(alphaSize, vars, maxConclusios, false);
			rules.put(vars[i], conclusios1);
		}
		return rules;
	}

	private List<List<String>> genConclusios(int alphaSize, String[] vars, int maxConclusios, boolean start) {
		List<List<String>> conclusios = new LinkedList<>();
		if (start && Math.random() <= 0.5) {
			conclusios.add(new LinkedList<>());
		}
		int numOfConclusios = random.nextInt(maxConclusios) + 1;
		while (conclusios.size() < numOfConclusios || ListOfListContains.contains(conclusios, new LinkedList<>())
				&& conclusios.size() < numOfConclusios + 1) {
			List<String> conclusio = genConclusio(alphaSize, vars);
			if (!ListOfListContains.contains(conclusios, conclusio)) {
				conclusios.add(conclusio);
			}
		}
		return conclusios;
	}

	private List<String> genConclusio(int alphaSize, String[] vars) {
		List<String> c = new LinkedList<>();
		String s = ((char) (random.nextInt(alphaSize) + 97)) + "";
		c.add(s);
		if (Math.random() < 0.7) {
			c.add(vars[random.nextInt(vars.length)]);
		}
		return c;
	}

	public List<Map<String, List<List<String>>>> genType3GrammarRuleSets0(String[] vars, int alphaSize,
			int maxConclusios, String s, int n) {
		List<Map<String, List<List<String>>>> grammarRuleSets = new LinkedList<>();
		while (grammarRuleSets.size() < n) {
			Map<String, List<List<String>>> rules = genType3GrammarRuleSet(vars, alphaSize, maxConclusios, s);
			grammarRuleSets.add(rules);
		}
		return grammarRuleSets;
	}

	public List<Map<String, List<List<String>>>> genType3GrammarRuleSets1(String[] vars, int alphaSize,
			int maxConclusios, String s, int n) {
		List<Map<String, List<List<String>>>> grammarRuleSets = new LinkedList<>();
		while (grammarRuleSets.size() < n) {
			Map<String, List<List<String>>> rules = genType3GrammarRuleSet(vars, alphaSize, maxConclusios, s);
			if (!CoReachability.emptyLanguage(rules, s)) {
				grammarRuleSets.add(rules);
			}
		}
		return grammarRuleSets;
	}

	public List<Map<String, List<List<String>>>> genType3GrammarRuleSets2(String[] vars, int alphaSize,
			int maxConclusios, String s, int n) {
		List<Map<String, List<List<String>>>> grammarRuleSets = new LinkedList<>();
		while (grammarRuleSets.size() < n) {
			Map<String, List<List<String>>> rules = genType3GrammarRuleSet(vars, alphaSize, maxConclusios, s);
			if (CoReachability.coReachableSet(rules).containsAll(rules.keySet())) {
				grammarRuleSets.add(rules);
			}
		}
		return grammarRuleSets;
	}

	public List<Map<String, List<List<String>>>> genType3GrammarRuleSets3(String[] vars, int alphaSize,
			int maxConclusios, String s, int n) {
		List<Map<String, List<List<String>>>> grammarRuleSets = new LinkedList<>();
		Set<String> alphaSet = new HashSet<>();
		for (int i = 0; i < alphaSize; i++) {
			alphaSet.add(((char) (i + 97)) + "");
		}
		while (grammarRuleSets.size() < n) {
			Map<String, List<List<String>>> rules = genType3GrammarRuleSet(vars, alphaSize, maxConclusios, s);
			boolean allCoReachable = CoReachability.coReachableSet(rules).containsAll(rules.keySet());
			boolean allReachable = Reachability.reachableSet(rules, s).containsAll(rules.keySet())
					&& Reachability.reachableSet(rules, s).containsAll(alphaSet);
			if (allCoReachable && allReachable) {
				grammarRuleSets.add(rules);
			}
		}
		return grammarRuleSets;
	}

}
